function [] = singleImageHazeRemoval(fileName, lambda, fileNameDarkChannel, fileNameTransmission, fileNameTransmissionOpt, fileNameOutput, patchSize)
% =========================================================================
%   Implementation and evaluation of paper: "Single Image Haze Removal
%   Using Dark Channel Prior"
%
%   Course final project: "Procesamiento de imágenes en color para cine y 
%   fotografía"
%   Author: Luis Di Martino
%   Date: Feb 2015
%
%   Usage:
%   The script assumes a folder called "Imagenes" exists in a upper path
%   level of the folder containing the code. This folder called "Imagenes"
%   is where input images should be located and results are saved. 
% =========================================================================

% clc
% clear all
% close all

% Parse input parameters
disp(['fileName : ' fileName])
disp(['lambda : ' lambda])
disp(['fileNameDarkChannel : ' fileNameDarkChannel])
disp(['fileNameTransmission : ' fileNameTransmission])
disp(['fileNameTransmissionOpt : ' fileNameTransmissionOpt])
disp(['fileNameOutput : ' fileNameOutput])
disp(['patchSize type : ' class(patchSize)])

% Parse input parameters
patchSize = str2num(patchSize);
lambda = str2num(lambda);

% Add paths
% addpath(strcat(pwd,filesep,'Display'));
% addpath(strcat(pwd,filesep,'LapMatrix',filesep,'learningBasedMatting'));

%% Script functioning
displayImages = 0;
atmLightProp = 0.001; % 0.1 percent

%% Definitions
% imPath = strcat(pwd,filesep,'..',filesep,'Imagenes',filesep);

imResizeScaleDarkChannel = 0.5;
imResizeScaleOptim = 0.2;
pixelsInputAtmLight = 10;
w = 0.80;
t0 = 0.1;
eps = 0.1;

%% 1) Image loading
inputIm = imread(fileName);
scaledInputIm = inputIm;
scaledInputIm = double(scaledInputIm);
if(displayImages)
    displayImage(scaledInputIm,'Input image');
end

disp('Image min size/patch size relation')
min(size(scaledInputIm,1)/patchSize)

tic


%% 2) Dark channel computation

% Compute minimum intensity values from RGB channels in the image.
minChannelsIm = min(scaledInputIm,[],3);
if(displayImages)
    displayImage(minChannelsIm,'Minimum rgb image');
end

% Minimum filter by patches in the minChannelsIm image
erosionWindow = strel('rectangle',[patchSize,patchSize]);
darkChannelIm = imerode(minChannelsIm,erosionWindow);
if(displayImages)
    displayImage(darkChannelIm,'Dark Channel');
end

%% 3) Estimation of the atmospheric light

% Order pixels in term of its brightness in the dark channel image
[sortedValues, sortedIndexs] = sort(darkChannelIm(:),'descend');
% Retain only the atmLightProp brighter pixels
sortedValues = sortedValues(1:round(size(sortedValues,1)*atmLightProp));
sortedIndexs = sortedIndexs(1:round(size(sortedIndexs,1)*atmLightProp));
[sortedIndexRows, sortedIndexCols] = ind2sub(size(darkChannelIm),sortedIndexs);
darkChannelAtmLightIm = repmat(darkChannelIm,[1 1 3]);
roiDarkChannel = zeros(size(darkChannelIm));

% Create image for the displaying of brightest pixels localization.
atmLightEstimationIm = scaledInputIm;

for index = 1:size(sortedIndexRows,1)
    roiDarkChannel(sortedIndexRows(index),sortedIndexCols(index)) = 1;
    
    atmLightEstimationIm(sortedIndexRows(index),sortedIndexCols(index),1) = 255;
    atmLightEstimationIm(sortedIndexRows(index),sortedIndexCols(index),2) = 0;
    atmLightEstimationIm(sortedIndexRows(index),sortedIndexCols(index),3) = 0;
    
    darkChannelAtmLightIm(sortedIndexRows(index),sortedIndexCols(index),1) = 255;
    darkChannelAtmLightIm(sortedIndexRows(index),sortedIndexCols(index),2) = 0;
    darkChannelAtmLightIm(sortedIndexRows(index),sortedIndexCols(index),3) = 0;
end

if(displayImages)
    displayImage(darkChannelAtmLightIm,'Dark channel brightest pixels');
    displayImage(roiDarkChannel,'ROI Dark Channel');
end

% Obtain highest intensity pixels from input image in the dark channel
% obtained ROI, consider only the 10 highest intensity pixels.
inputBWIm = rgb2gray(scaledInputIm);
inputBWIm = inputBWIm.*roiDarkChannel;
[sortedValuesInputBW, sortedIndexsInputBW] = sort(inputBWIm(:),'descend');
[sortedIndexRowsInputIm, sortedIndexColsInputIm] = ind2sub(size(inputBWIm),sortedIndexsInputBW(1:pixelsInputAtmLight));
maxIntensityMask = zeros(size(scaledInputIm,1),size(scaledInputIm,2),1);

for index = 1:size(sortedIndexRowsInputIm,1)
    maxIntensityMask(sortedIndexRowsInputIm(index),sortedIndexColsInputIm(index)) = 1;
end

% Filter each channel and take the mean over highest intensity pixels
inputImFiltered_R = scaledInputIm(:,:,1).*maxIntensityMask(:,:);
inputImFiltered_G = scaledInputIm(:,:,2).*maxIntensityMask(:,:);
inputImFiltered_B = scaledInputIm(:,:,3).*maxIntensityMask(:,:);

atmLightR = floor(sum(sum(inputImFiltered_R))/pixelsInputAtmLight);
atmLightG = floor(sum(sum(inputImFiltered_G))/pixelsInputAtmLight);
atmLightB = floor(sum(sum(inputImFiltered_B))/pixelsInputAtmLight);

%% 4) Computing the transmission without using soft matting

% Divide each channel by the atmLight
normHazyIm(:,:,1) = double(scaledInputIm(:,:,1))/atmLightR;
normHazyIm(:,:,2) = double(scaledInputIm(:,:,2))/atmLightG;
normHazyIm(:,:,3) = double(scaledInputIm(:,:,3))/atmLightB;

% Compute dark channel of the normalized hazy image.
minChannelsNormHazyIm = min(normHazyIm,[],3);
if(displayImages)
    displayImage(minChannelsNormHazyIm,'Minimum normalized hazy image');
end

% Minimum filter by patches in the minChannelsIm image
erosionWindow = strel('rectangle',[patchSize,patchSize]);
darkChannelNormHazyIm = imerode(minChannelsNormHazyIm,erosionWindow);

if(displayImages)
    displayImage(darkChannelNormHazyIm,'Dark channel of normalized hazy image');
end

% Compute transmission according to the dark channel of the normalized
% hazy image.
transIm = 1-w.*darkChannelNormHazyIm;
if(displayImages)
    displayImage(transIm,'Transmission without soft matting');
end


%% 4.d) Compute Laplacian Matrix using demo code
optimScaledInputIm = imresize(inputIm,[imResizeScaleOptim*size(inputIm,1) imResizeScaleOptim*size(inputIm,2)]);
L = getLap_iccv09_overlapping(optimScaledInputIm,[3 3], zeros(size(optimScaledInputIm,1),size(optimScaledInputIm,2)));

%% 5) Solve optimal t using L and the previous transmission map.
t_tilde = imresize(transIm,[imResizeScaleOptim*size(inputIm,1) imResizeScaleOptim*size(inputIm,2)]);
t_tilde = reshape(t_tilde,[],1);
t_optimal = (L+lambda*sparse(1:size(L,1),1:size(L,1),1))\(lambda*t_tilde);
%%
transOptIm = reshape(t_optimal,size(optimScaledInputIm,1),size(optimScaledInputIm,2));
transOptIm = imresize(transOptIm,[size(inputIm,1) size(inputIm,2)],'bilinear');


%% 6) Recover scene radiance
jAuxIm = double(scaledInputIm);
atmLightIm = zeros(size(scaledInputIm));
atmLightIm(:,:,1) = atmLightR;
atmLightIm(:,:,2) = atmLightG;
atmLightIm(:,:,3) = atmLightB;

radianceIm = (jAuxIm-atmLightIm)./repmat(max(t0,transIm),[1 1 3])+atmLightIm;
radianceIm = uint8(radianceIm);

if(displayImages)
    displayImage(radianceIm,'Radiance of original scene');
end

%% 7) Recover scene radiance using t_optimal

radianceImToptimal = (jAuxIm-atmLightIm)./repmat(max(t0,transOptIm),[1 1 3])+atmLightIm;
radianceImToptimal = uint8(radianceImToptimal);

if(displayImages)
    displayImage(radianceIm,'Radiance of original scene');
end

toc

%% 8) Convert double to images
transIm = uint8(255*transIm);
transOptIm = uint8(255*transOptIm);
scaledInputIm = uint8(scaledInputIm);
darkChannelIm = uint8(darkChannelIm);
atmLightEstimationIm = uint8(atmLightEstimationIm);

%% 9) Save result

imwrite(darkChannelIm, fileNameDarkChannel)
imwrite(transIm, fileNameTransmission)
imwrite(transOptIm, fileNameTransmissionOpt)
imwrite(radianceImToptimal, fileNameOutput)
% imwrite(radianceIm,strcat(imPath,filesep,fileNameResult,'.jpg'));
% imwrite(radianceImToptimal,strcat(imPath,filesep,fileNameResultOpt,'.jpg'));
% imwrite(transIm,strcat(imPath,filesep,fileNameTransIm,'.jpg'));
% imwrite(transOptIm,strcat(imPath,filesep,fileNameTransOptIm,'.jpg'));
% imwrite(scaledInputIm,strcat(imPath,filesep,fileNameOriginal,'.jpg'));
% imwrite(darkChannelIm,strcat(imPath,filesep,fileNameDarkChannel,'.jpg'));
% imwrite(atmLightEstimationIm,strcat(imPath,filesep,fileNameAtmLightIm,'.jpg'));
%imwrite(radianceImToptimal,fileNameResultOpt);

